/* fed.c - field editing package */

/*  Copyright 1991 Mark Russell, University of Kent at Canterbury.
 *
 *  You can do what you like with this source code as long as
 *  you don't try to make money out of it and you include an
 *  unaltered copy of this message (including the copyright).
 */

/*  The routines here are meant to make field editing easier.
 */

char shared_fed_c_sccsid[] = "@(#)fed.c	1.21 26/7/92 (UKC)";

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include <local/wn.h>

#include <local/ukcprog.h>

#include "fed.h"
#include "obj.h"
#include "o_priv.h"

typedef struct popupst {
	int po_mdesc;		/* Menu descriptor */
	int po_save_last;	/* If set, remember caption the user last selected */
	int po_last;		/* Last caption selected by user */
	const char **po_caps;	/* NULL terminated array of captions */
} popup_t;

static void get_act PROTO((struct edescst *edesc));
static void preprocess_key PROTO((struct edescst *edesc));
static void preprocess_puck PROTO((struct edescst *edesc));
static int cancel_edit PROTO((struct edescst *edesc));
static void paste_selection PROTO((edesc_t *edesc));
static void paste_function_key_string PROTO((edesc_t *edesc));
static void redraw_string PROTO((edesc_t *edesc));
static char *make_copy PROTO((const char *s, int len));
static dump_obj PROTO((char *arg, int level, char *line));

void dump_selected_objects PROTO((void));

/*  Event mask for wn_next_event.
 */
#define EVENT_MASK	(EV_BUTTON_UP | EV_BUTTON_DOWN | EV_MOUSE_MOVED | \
					EV_KEY | EV_INTERRUPT | \
					EV_WINDOW_EXPOSED | EV_WINDOW_RESIZED)

static int
dump_obj(arg, level, line)
char *arg;
int level;
char *line;
{
	fprintf(stderr, "\t%*s%s\n", level * 4, "", line);

	/*  The format_obj callbacks all use strf() to create the line.
	 */
	free(line);
	
	return 0;
}

static char *
make_copy(s, len)
const char *s;
int len;
{
	return strncpy(e_malloc(len + 1), s, len);
}

/*  Initialise struct edescst. Set ed_{wn,orig,maxlen,scrlen}
 *  to the given values. Make a copy of orig in ed_copy, and set
 *  ed_curlen to the length of orig.
 */
void
make_edesc(edesc, wn, orig, maxlen, fg, bg)
register struct edescst *edesc;
window_t wn;
const char *orig;
int maxlen, fg, bg;
{
	edesc->ed_wn = wn;
	edesc->ed_font = (wn == -1) ? NULL : wn_get_sysfont();
	edesc->ed_fg = fg;
	edesc->ed_bg = bg;
	edesc->ed_orig = orig;
	edesc->ed_maxlen = maxlen;
	edesc->ed_scrlen = 0;
	edesc->ed_flags = 0; /* cursor off, line cursor, insert mode */
	edesc->ed_copy = make_copy(orig, maxlen);
	edesc->ed_copy[maxlen] = '\0';
	edesc->ed_curlen = strlen(edesc->ed_copy);
	edesc->ed_quitfunc = NULL;
	edesc->ed_keyfunc = NULL;
	edesc->ed_puckfunc = NULL;
	edesc->ed_expose_handler = NULL;
}

/*  Get the next user action using wn_next_event().
 *  Set the fields ed_{act,char,x,y,buttons} of *edesc appropriately.
 */
static void
get_act(edesc)
register struct edescst *edesc;
{
	event_t event;

	wn_next_event(edesc->ed_wn, EVENT_MASK, &event);
	if (event.ev_type == EV_KEY) {
		edesc->ed_act = EDA_KEY;
		edesc->ed_char = event.ev_char;
	}
	else if ((event.ev_type & (EV_WINDOW_RESIZED | EV_WINDOW_EXPOSED)) ||
		 event.ev_buttons & B_ANY) {
		edesc->ed_act = EDA_PUCK;
		edesc->ed_puck_x = event.ev_x;
		edesc->ed_puck_y = event.ev_y;
		edesc->ed_buttons = event.ev_buttons |
				    (event.ev_type & (EV_BUTTON_UP|EV_BUTTON_DOWN));
	}
	else
		edesc->ed_act = EDA_NONE;
}

/*  Do the default processing for key ch. Interprets the characters
 *  RETURN (confirm edit), DEL (delete last character), and OOPS
 *  (delete to start of line).
 *  If the character is any of these, return EDM_INSERT_CHAR if it
 *  is printable, otherwise EDM_ILLEGAL_CHAR.
 */
static void
  preprocess_key(edesc)
register struct edescst *edesc;
{
  int meaning;
  
  switch(edesc->ed_char) {
  case RETURN:
  case ESC:
    edesc->ed_flags &= ~EDF_UNGETPUCK;
    meaning = EDM_CONFIRM;
    break;

  /* Make Control-D work like Bashers expect */
  case CONTROL_D:
    meaning = EDM_DELETE_NEXT;
    break;

  /*  Delete to white space, like bash */
  case CONTROL_W:
    meaning = EDM_DELETE_TO_WHITE;
    break;

  /* Delete to start of indented line, like vi d^ */
  case CONTROL_CARAT:
    meaning = EDM_DELETE_TO_LINE_START;
    break;

  /* Delete to start of C token, like vi db */
  case CONTROL_T:
    meaning = EDM_DELETE_C_TOKEN;
    break;


  case BACKSPACE:		/* RGA moved from CONTROL_B */
  case DEL:
    meaning = EDM_DELETE_CHAR;
    break;
  case CONTROL_A:
    meaning = EDM_BEGINNING_OF_LINE;
    break;
  case OOPS:
    meaning = EDM_DELETE_LINE;
    break;
  case CONTROL_C:
    meaning = EDM_CANCEL;
    break;
/* RGA this it obseleted by AMC stuff above */
/*  case CONTROL_D:
    meaning = EDM_DELETE_THIS_CHAR;
    break;*/
  case CONTROL_E:
    meaning = EDM_END_OF_LINE;
    break;
  case CONTROL_K:
    meaning = EDM_DELETE_REST_OF_LINE;
    break;
  case CONTROL_L:
  case CONTROL_F:
  case RIGHT_ARROW:
    meaning = EDM_FORWARD_SPACE;
    break;
  case CONTROL_B:
  case LEFT_ARROW:
    meaning = EDM_BACK_SPACE;
    break;
  case CONTROL_O:
    meaning = EDM_CONT;
    if (edesc->ed_flags & EDF_OVERWRITE)
    {
      edesc->ed_flags &= ~EDF_OVERWRITE;
      errf("\binsert editing");
    }
    else
    {
      edesc->ed_flags |= EDF_OVERWRITE;
      errf("\boverwrite editing");
    }
    break;
  case CONTROL_S:
    meaning = EDM_SEARCH_DUMP;
    break;
  case CONTROL_Y:
  case UP_ARROW:
    meaning = EDM_PASTE_SELECTION;
    break;
  case F1_KEY:
  case F2_KEY:
  case F3_KEY:
  case F4_KEY:
  case F5_KEY:
  case F6_KEY:
  case F7_KEY:
  case F8_KEY:
  case F9_KEY:
  case F10_KEY:
  case F11_KEY:
  case F12_KEY:
    meaning = EDM_INSERT_FUNC_STR;
    break;
  default:
    meaning = (edesc->ed_char < 32) ? EDM_ILLEGAL_CHAR
      : EDM_INSERT_CHAR;
    break;
  }
  edesc->ed_meaning = meaning;
}


/*  Do default processing for the puck. B_MIDDLE indicates new text
 *  cursor position. A press of B_MIDDLE outside the field area or
 *  a press of B_RIGHT or B_LEFT anywhere means end of edit.
 *
 *  x,y, and buttons are the values of edesc->ed_{x,y,buttons},
 */
static void
preprocess_puck(edesc)
register struct edescst *edesc;
{
#define ROUNDI(f) (((f) >=0.0) ? (int)((f)+0.5): (int)((f)-0.5))
#define min(x,y) ((x) > (y) ? (y) : (x))
#define max(x,y) ((x) > (y) ? (x) : (y))

	int x, y, buttons;
	unsigned long resize_type;
	
	x = edesc->ed_puck_x;
	y = edesc->ed_puck_y;
	buttons = edesc->ed_buttons & (B_ANY | B_SHIFT_KEY);
	resize_type = wn_get_resize_event(edesc->ed_wn);
	if (resize_type == EV_WINDOW_EXPOSED && edesc->ed_expose_handler != NULL) {
		ed_cursor(edesc, EDC_CURSOR_OFF);
		(*edesc->ed_expose_handler)(resize_type, TRUE);
		redraw_string(edesc);
		edesc->ed_meaning = EDM_CONT;
	}
	else if (resize_type != 0) {
		event_t event;

		edesc->ed_meaning = EDM_CANCEL;

		/*  We just consumed the resize event, so put it back.
		 */
		event.ev_wn = edesc->ed_wn;
		event.ev_type = resize_type;
		event.ev_x = x;
		event.ev_y = y;
		wn_pushback_event(&event);
	}
	else if (buttons == (B_MIDDLE | B_SHIFT_KEY)) {
		edesc->ed_meaning = EDM_PASTE_SELECTION;
		wn_wait_for_release_of(edesc->ed_wn, B_MIDDLE);
	}
	else if (buttons == B_LEFT &&
		 edesc->ed_buttons & B_CONTROL_KEY)
	{
	  char buff[256];
	  int last_col, col;
	  event_t ev;

	  last_col = ROUNDI(edesc->ed_puck_x/edesc->ed_font->ft_width);
	  last_col = max(0, min(strlen(edesc->ed_copy), last_col));

	  edesc->ed_meaning = EDM_CONT;
	  wn_get_release_event(edesc->ed_wn, buttons, &ev);
	  set_event_win(&ev);
	  if (edesc->ed_wn == ev.ev_wn)
	  {
	    col = ROUNDI(ev.ev_x/edesc->ed_font->ft_width);
	    col = max(0, min(strlen(edesc->ed_copy), col));
	    
	    buff[0] = 0;
	    strncat(buff, edesc->ed_copy + min(last_col, col),
		    last_col <= col ? col - last_col : last_col - col);
	    wn_set_selection(buff, strlen(buff));
	  }
	}
	else if (buttons == B_RIGHT) {
		static const char *bptcaps_strs[] = {
			"UPS_F1_STR",
			"UPS_F2_STR",
			"UPS_F3_STR",
			"UPS_F4_STR",
			"UPS_F5_STR",
			"UPS_F6_STR",
			"UPS_F7_STR",
			"UPS_F8_STR",
			"UPS_F9_STR",
			"UPS_F10_STR",
			"UPS_F11_STR",
			"UPS_F12_STR",
			NULL
		};
		static const char *bptcaps[13];
		static const char key_chars[] = {
		 F1_KEY, F2_KEY, F3_KEY, F4_KEY, F5_KEY, F6_KEY,
		 F7_KEY, F8_KEY, F9_KEY, F10_KEY, F11_KEY, F12_KEY};
		static i = -1;
		static char ed_chars[12];
		char *env;
		static popup_t bptpop = { -1, FALSE, 0, bptcaps };
		int res, j;

		if (i == -1)
		{
		  for (i = 0, j = 0;j < 12 ; j++)
		  {
		    env = getenv(bptcaps_strs[j]);
		    if (env)
		    {
		      char *copy;
		      int controls = 0;

		      ed_chars[i] = key_chars[j];

		      /*
		      **   Count the number of control characters in env
		      */
		      for ( copy = env; *copy; copy++ )
			if ( *copy < 0x20) controls++;

		      /*
		      **   Now allocate space for the string
		      **   and copy, mapping Control-X to ^X.
		      */
		      bptcaps[i] = copy = malloc(strlen(env)+1 + controls);
		      for ( ; *env; env++ )
		      {
			if ( *env < 0x20 )
			{
			  *copy++ = '^';
			  *copy++ = 0x40 | *env;
			}
			else
			  *copy++ = *env;
		      }
		      *copy='\0';
		      i++;
		    }
		  }
		  bptcaps[i] = NULL;
		}
		if (i > 0)
		{
		  res = select_from_popup(edesc->ed_wn, B_RIGHT, &bptpop,
					  edesc->ed_puck_x, edesc->ed_puck_y);
		  switch(res) {
		  case 0:
		  case 1: 
		  case 2: 
		  case 3: 
		  case 4: 
		  case 5: 
		  case 6: 
		  case 7: 
		  case 8: 
		  case 9: 
		  case 10:
		  case 11:
		    edesc->ed_meaning = EDM_INSERT_FUNC_STR;
		    edesc->ed_char = ed_chars[res];
		    break;
		  default:
		    break;
		  }
		}
	      }
	else if (buttons & B_LEFT ||
			((edesc->ed_buttons & B_STATE_CHANGE) &&
			       (x < 0 || x >= wn_get_width(edesc->ed_wn) ||
				y < 0 || y >= wn_get_height(edesc->ed_wn)))) {
		edesc->ed_flags |= EDF_UNGETPUCK;
		edesc->ed_meaning = EDM_CONFIRM;
	}
	else if (buttons == B_MIDDLE) {
		edesc->ed_meaning = EDM_SETCURSOR;
		edesc->ed_newpos = wn_strpos(edesc->ed_copy, x, edesc->ed_font,
				       !(edesc->ed_flags & EDF_BLOCK_CURSOR));
	}
	else
		panic("no buttons pressed in preprocess_puck");
}

/*  For setting and getting display win search state
 */
int
display_win_search(set, reset)
     int set;
     int reset;
{
  static int search;

  if (set)
    search = 1;
  if (reset)
    search = 0;
  return(search);
    
}

/*  Do the edit indicated by the ed_meaning field of edesc
 */
void
  do_edit(edesc, meaning)
register struct edescst *edesc;
int meaning;
{
  int ch, old_state;
  
  switch(meaning) {
  case EDM_ILLEGAL_CHAR:
    ch = edesc->ed_char;
    if (isascii(ch) && isprint(ch))
      errf("illegal character");
    else if (ch >= 0 && ch < 32)
      errf("illegal character ^%c", ch + 64);
    else
      errf("illegal character <%d>", ch);
    break;
  case EDM_SETCURSOR:
    /*  move editing cursor to ed_newpos */
    if (edesc->ed_newpos > edesc->ed_curlen)
      edesc->ed_newpos = edesc->ed_curlen;
    else if (edesc->ed_newpos < 0)
      edesc->ed_newpos = 0;
    ed_cursor(edesc, EDC_CURSOR_OFF);
    edesc->ed_curpos = edesc->ed_newpos;
    ed_cursor(edesc, EDC_CURSOR_ON);
    break;
  case EDM_PASTE_SELECTION:
    paste_selection(edesc);
    break;
  case EDM_INSERT_FUNC_STR:
    paste_function_key_string(edesc);
    break;
  case EDM_DUMP_OBJ:
    dump_selected_objects();
    meaning = EDM_DUMP_OBJ;
    break;
  case EDM_DUMP_WIN:
    dump_object((objid_t)NULL, dump_obj, (char *)stdout, OBJ_DESCENDENTS);
    meaning = EDM_DUMP_WIN;
    break;
  case EDM_SEARCH_DUMP:
    display_win_search(1, 0);	/* set */
    old_state = updating(OBJ_UPDATING_QUERY);
    if (old_state == OBJ_UPDATING_OFF)
      updating(OBJ_UPDATING_ON);
    open_output_win();
    updating(old_state);
    dump_object((objid_t)NULL, dump_obj, (char *)stdout, OBJ_DESCENDENTS);
    display_win_search(0, 1);	/* reset */
    break;
  default:
    process_key(edesc, meaning);
    break;
  }
}

static void
paste_function_key_string(edesc)
edesc_t *edesc;
{
	const char *buf, *str;
	int i;
	/*  
	**   Allow control chars in paste string.
	*/
	int escape = FALSE;
	int control = FALSE;

	switch(edesc->ed_char) 
	{
	case F1_KEY: str = "UPS_F1_STR"; break;
	case F2_KEY: str = "UPS_F2_STR"; break;
	case F3_KEY: str = "UPS_F3_STR"; break;
	case F4_KEY: str = "UPS_F4_STR"; break;
	case F5_KEY: str = "UPS_F5_STR"; break;
	case F6_KEY: str = "UPS_F6_STR"; break;
	case F7_KEY: str = "UPS_F7_STR"; break;
	case F8_KEY: str = "UPS_F8_STR"; break;
	case F9_KEY: str = "UPS_F9_STR"; break;
	case F10_KEY: str = "UPS_F10_STR"; break;
	case F11_KEY: str = "UPS_F11_STR"; break;
	case F12_KEY: str = "UPS_F12_STR"; break;
	default: return;
	}
	buf = getenv(str);
	if (!buf)
	{
	  errf("No paste: environment variable %s undefined", str);
	  return;
	}

	for (i = 0; i < strlen(buf); ++i) {
		edesc->ed_char = buf[i];

		if ( !escape )
		{
		    if ( control )
		    {
		        edesc->ed_char &= 0x1F;
			control = FALSE;
		    }
		    else if (edesc->ed_char == '\\' )
		    {
			escape = TRUE;
			continue;
		    }
		    else if (edesc->ed_char == '^' )
		    {
			control = TRUE;
			continue;
		    }
		} else
		{
		     switch ( edesc->ed_char )
		     {
		     case '^':
		     case '\\':
			 /*
			 **   Pass \^ or \\ as ^ or \ respectively
			 */
			 break;

		     case 'n':
			/*
			** We paste in ESC rather than RETURN
			** here because of the check for RETURN
			** in the default case in edit_field
			*/
		     case 'e':
			/*
			**   If at the end of the buffer, map to ESC to terminate the edit
			*/
			if ( buf[i+1] == '\0' )
			{
			    edesc->ed_char = ESC;
			    break;
			}
			/* else fall into default case */
		     default:
			/*
			**    Back up and pass the '\' through
			*/
			edesc->ed_char = '\\';
			i--;
		     }
		     escape = FALSE;	/* New */
		}

		preprocess_key(edesc);
		if (edesc->ed_keyfunc != NULL)
			(*edesc->ed_keyfunc)(edesc);

		if (   edesc->ed_meaning == EDM_CONFIRM
		    || edesc->ed_meaning == EDM_CANCEL )
		    break;

		if (   edesc->ed_meaning == EDM_CONT )
		    continue;

		/*
		**   We ignore error codes from process_key
		**   because to exit on them causes the
		**   user to drop out if, for example the pasted script
		**   begins with ^u and the edit area is empty.
		**    Call do_edit instead of process_key, since this handles ^y
		**    and illegal chars.
		*/
		do_edit(edesc, edesc->ed_meaning);
		if ( edesc->ed_meaning == EDM_ILLEGAL_CHAR )
		    break;
	}
}

static void
  paste_selection(edesc)
edesc_t *edesc;
{
  const char *buf;
  int i, nbytes, new_line = 0;
typedef enum { QR_CONFIRM, QR_NEWLINE, QR_DEL, QR_OTHER } quit_reason_t;
  
  wn_get_selection(&buf, &nbytes);
  
  for (i = 0; i < nbytes; ++i)
  {
    if (buf[i] == 10)	/* return */
    {
      new_line++;
      edesc->ed_char = RETURN;
      edesc->ed_user = QR_NEWLINE;
    }
    if (new_line)
      _wn_send_back_event(buf[i], new_line == 1);
    else
    {
      edesc->ed_char = buf[i];
      
      preprocess_key(edesc);
      if (edesc->ed_keyfunc != NULL)
	(*edesc->ed_keyfunc)(edesc);
      
      if (edesc->ed_meaning != EDM_INSERT_CHAR)
      {
	wn_bell(edesc->ed_wn);
	break;
      }
      
      if (process_key(edesc, edesc->ed_meaning) != 0)
	break;
    }
  }
}

/*  Field editing routine. This is called to edit a field, and
 *  only returns when the edit is confirmed or cancelled (except
 *  when the EDF_ONETRIP bit is set - see below).
 *
 *  We go round in a loop, calling get_act() to get a user action,
 *  which can be a puck press or a key press. The raw input record
 *  is interpreted by preprocess_puck() or preprocess_key(), then
 *  passed to the ed_key() or ed_puck() function given in edesc so
 *  that the user can alter the default interpretation.
 *  At this point, the meaning of the action is recorded in the
 *  ed_meaning field, and do_edit() is called to perform the edit.
 *  unless the meaning is EDM_CONT, EDM_CANCEL or EDM_CONFIRM.  
 *  ... more comment needed
 */
int
edit_field(edesc, expand_on_tab)
register struct edescst *edesc;
int expand_on_tab;
{
	int n_quit_attempts, meaning, res, quit;
	int dump_to_out_win = 0, old_state, offset;
	
	n_quit_attempts = 0;
	quit = FALSE;
	do {
		if (!(edesc->ed_flags & EDF_ONETRIP))
			do {
				get_act(edesc);
			} while (edesc->ed_act == EDA_NONE);
		if (expand_on_tab && edesc->ed_char == CONTROL('I')) /* tab */
		{
		  dump_to_out_win = 1;
		  old_state = updating(OBJ_UPDATING_QUERY);
		  if (old_state == OBJ_UPDATING_OFF)
		    updating(OBJ_UPDATING_ON);
		  open_output_win();
		  if (edesc != (struct edescst *)get_typing_line_edesc())
		  {
		    offset = ensure_visible((objid_t)edesc->ed_user);
		    if (offset)
		      wn_update_y_offs(edesc->ed_wn, -offset);
		  }
		  updating(old_state);
		  ed_cursor((struct edescst *)
			    get_typing_line_edesc(), EDC_CURSOR_OFF);	
		  ed_cursor(edesc, EDC_CURSOR_OFF);
		  redraw_string(edesc);
		  edesc->ed_char = '\033';	/* escape */
		  display_win_search(1, 0);	/* set */
		}
		if (edesc->ed_act == EDA_KEY) {
		  preprocess_key(edesc);
		  if (edesc->ed_keyfunc != NULL)
		    (*edesc->ed_keyfunc)(edesc);
		}
		else {
			preprocess_puck(edesc);
			if (edesc->ed_puckfunc != NULL)
				(*edesc->ed_puckfunc)(edesc);
		}
		switch(meaning = edesc->ed_meaning) {
		case EDM_CONFIRM:
    confirm:
		  ed_cursor(edesc, EDC_CURSOR_OFF);
			res = (*edesc->ed_quitfunc)(edesc, n_quit_attempts);
			if (dump_to_out_win)
			{
			  dump_to_out_win = 0;
			  if (res != EDR_CONFIRM_CHANGE)
			    res = EDR_CANT_QUIT_EXP;
			  display_win_search(0, 1);	/* reset */
			}
			if (res == EDR_CANT_QUIT_EXP || res == EDR_CONT)
			{
			  if (res == EDR_CANT_QUIT_EXP)
			    res = EDR_CANT_QUIT;
			  edesc->ed_curpos =
			    edesc->ed_curlen = strlen(edesc->ed_copy);
			  wn_text(edesc->ed_wn, edesc->ed_font, edesc->ed_copy,
				  0, 0, edesc->ed_fg, edesc->ed_bg, WN_USE_TOP);
			}
			if (res == EDR_CANT_QUIT)
				n_quit_attempts++;
			quit =  res == EDR_CONFIRM_CHANGE ||
				res == EDR_CONFIRM_NO_CHANGE ||
				res == EDR_CANCEL_AND_QUIT;
			if (!quit)
				ed_cursor(edesc, EDC_CURSOR_ON);
			break;
		case EDM_CONT:
			res = EDR_CONT;
			break;
		case EDM_CANCEL:
			res = EDR_CANCEL;
			break;
		default:
			do_edit(edesc, meaning);
			n_quit_attempts = 0;
			res = EDR_CONT;
			if (edesc->ed_char == RETURN)
			{
			  res = EDR_CONFIRM_CHANGE;
			  return res;
			}
			if (edesc->ed_meaning == EDM_CONFIRM) 
			    goto confirm;
			break;
		}
		if (meaning == EDM_CANCEL || res == EDR_CANCEL ||
					     res == EDR_CANCEL_AND_QUIT)
			quit = cancel_edit(edesc);
	} while (!(quit || (edesc->ed_flags & EDF_ONETRIP)));
	
	/*  unget the puck record if the input record can be
	 *  used by the next process to call wn_getpuck.
	 */
	if (quit && (edesc->ed_flags & EDF_UNGETPUCK))
		wn_ungetpuck(edesc->ed_wn, edesc->ed_puck_x,
					edesc->ed_puck_y, edesc->ed_buttons);
	return res;
}

/*  Replace ed_orig with new_orig.
 */
void
ed_new_orig(edesc, new_orig)
struct edescst *edesc;
const char *new_orig;
{
	edesc->ed_orig = new_orig;
}

/*  Replace the string we are editing with s, and put the cursor
 *  at the position indicated by newpos.
 */
void
ed_newstring(edesc, s)
struct edescst *edesc;
const char *s;
{
	ed_cursor(edesc, EDC_CURSOR_OFF);
	if (edesc->ed_copy != NULL)
		free(edesc->ed_copy);
	edesc->ed_copy = make_copy(s, edesc->ed_maxlen);
	edesc->ed_curlen = strlen(s);
	redraw_string(edesc);
}			

static void
redraw_string(edesc)
edesc_t *edesc;
{
	int wn;

	edesc->ed_newpos = edesc->ed_curpos;

	/*  Sometimes ups wants to run without a window ...
	 */
	if ((wn = edesc->ed_wn) == -1)
		return;

	wn_set_area(wn, 0, 0, wn_get_width(wn), wn_get_height(wn), WN_BG);
	wn_text(wn, edesc->ed_font, edesc->ed_copy,
		0, 0, edesc->ed_fg, edesc->ed_bg, WN_USE_TOP);
	do_edit(edesc, EDM_SETCURSOR);
}

/*  Cancel an edit. Restore the original string on screen, update the
 *  working copy from the original string, and ensure that the text
 *  cursor is still in bounds (the original string may have been shorter
 *  than the cancelled buffer).
 *  Return whether we should finish the edit
 */
static int
cancel_edit(edesc)
register struct edescst *edesc;
{
	int oldwidth, newwidth, height;
	font_t *font;
	window_t wn;
	
	ed_cursor(edesc, EDC_CURSOR_OFF);
	wn = edesc->ed_wn;
	font = edesc->ed_font;
	oldwidth = wn_strwidth(edesc->ed_copy, font);
	newwidth = wn_strwidth(edesc->ed_orig, font);
	height = font->ft_height;
	wn_set_area(wn, 0, 0, oldwidth, height, WN_BG);
	if (edesc->ed_scrlen != 0)
		wn_move_area(wn, oldwidth, 0, edesc->ed_scrlen, height,
							newwidth, 0, WN_BG);
	wn_text(wn, font, edesc->ed_orig, 0, 0, edesc->ed_fg, edesc->ed_bg, WN_USE_TOP);
	(void) strcpy(edesc->ed_copy, edesc->ed_orig);
	edesc->ed_curlen = strlen(edesc->ed_copy);
	if (edesc->ed_curpos > edesc->ed_curlen)
		edesc->ed_curpos = edesc->ed_curlen;
	if (edesc->ed_flags & EDF_CONT_ON_CANCEL) {
		ed_cursor(edesc, EDC_CURSOR_ON);
		return FALSE;
	}
	return TRUE;
}

/*  Process EDM_DELETE_LINE, EDM_DELETE_CHAR, EDM_INSERT_CHAR,
 *  , EDM_DELETE_NEXT, EDM_FORWARD_SPACE, and EDM_BACK_SPACE.
 *
 *  Call smove() to update the working copy, and ed_cursor() to update
 *  the text cursor, as necessary.
 *
 *  Beep for DEL/OOPS at the start of a line, and normal characters
 *  when we have reached the maximum buffer length
 */
int
process_key(edesc, meaning)
register struct edescst *edesc;
int meaning;
{
	window_t wn;
	int pos, newpos, newlen, do_move, ch, overwrite;
	font_t *font;
	int old_x, new_x, width, height, j;
	char *s, buf[2];
	
	pos = edesc->ed_curpos;
	ch = edesc->ed_char;
	switch(meaning) {
	case EDM_DELETE_LINE:
		newpos = (pos == 0) ? -1 : 0;
		do_move = TRUE;
		break;
	case EDM_INSERT_CHAR:
		newpos = pos + 1;
		do_move = TRUE;
		break;
	case EDM_FORWARD_SPACE:
		newpos = pos + 1;
		do_move = FALSE;
		break;
	case EDM_BEGINNING_OF_LINE:
		newpos = 0;
		do_move = FALSE;
		break;
	case EDM_END_OF_LINE:
		newpos = edesc->ed_curlen;
		do_move = FALSE;
		break;
	case EDM_BACK_SPACE:
	case EDM_DELETE_CHAR:
		newpos = pos - 1;
		do_move = meaning == EDM_DELETE_CHAR;
		break;
/* RGA this it obseleted by AMC stuff below */
/*	case EDM_DELETE_THIS_CHAR:
		newpos = pos;
		if (pos < edesc->ed_maxlen)
		  pos++;
		do_move = TRUE;
		break;*/

	/*   Implement Control-D for delete */
	case EDM_DELETE_NEXT:
		newpos = pos++;
		do_move = 1;
		break;

	/*   Implement Control-W  */
	case EDM_DELETE_TO_WHITE:
		newpos = pos;
		s = edesc->ed_copy+pos;
		while (newpos > 0 && *--s == ' ')
		    newpos--;
		while (newpos > 0 && *s-- != ' ')
		    newpos--;
		do_move = 1;
		break;

	/*   Implement Control-^  */
	case EDM_DELETE_TO_LINE_START:
		s = edesc->ed_copy;
		newpos = 0;
		while(   newpos < pos
		      && *s++ == ' ' )
		    newpos++;
		do_move = 1;
		break;

	/*   Implement Control-T  */
	case EDM_DELETE_C_TOKEN:
		newpos = pos;
		if (newpos == 0)
		   break;
		s = edesc->ed_copy+pos;
		if (*--s == '_' || isalnum(*s) )
		{
		    /* Delete to start of C identifer */
		    do newpos--;
		    while ( newpos > 0
			   && (*--s == '_' || isalnum(*s) ));
		} else if (*s > ' ')
		{
		    /* Delete to start of other C token */
		    do newpos--;
		    while (   newpos > 0
			   && ( *--s != ' ' )
			   && !( isalnum(*s) || *s == '_' ) );
		} else
		{
		    /* Delete to end of C token */
		    do newpos--;
		    while (   newpos > 0
			   && *--s == ' ' );
		}
		do_move = 1;
		break;
	case EDM_DELETE_REST_OF_LINE:
		s = edesc->ed_copy;
		font = edesc->ed_font;
		for (j = pos; j < edesc->ed_curlen; j++)
		  s[j] = ' ';
		s[j] = 0;
		wn_updating_off(edesc->ed_wn);
		ed_cursor(edesc, EDC_CURSOR_OFF);
		wn_text(edesc->ed_wn, font, edesc->ed_copy, 0, 0,
			edesc->ed_fg, edesc->ed_bg, WN_USE_TOP);
		edesc->ed_copy[pos] = 0;
		newpos = pos;
		do_move = FALSE;
		edesc->ed_curlen = newpos;
		ed_cursor(edesc, EDC_CURSOR_ON);
		wn_updating_on(edesc->ed_wn);
		return 0;
	default:
		panic("unknown meaning type in process_key");
		newpos = 0; /* to satisfy gcc */
		do_move = 0; /* to satisfy gcc */
	}

	overwrite = (edesc->ed_flags & EDF_OVERWRITE) ? 1 : 0;
	newlen = edesc->ed_curlen;

	if (do_move && !overwrite)
		newlen += newpos - pos;

	if (newpos < 0 || newpos > newlen || newlen > edesc->ed_maxlen) {
		wn_bell(edesc->ed_wn);
		return -1;
	}

	s = edesc->ed_copy;
	font = edesc->ed_font;
	wn = edesc->ed_wn;
	height = font->ft_height;

	wn_updating_off(wn);
	ed_cursor(edesc, EDC_CURSOR_OFF);
	old_x = wn_strnwidth(s, pos, font);

	if (do_move && !overwrite) {
		int i;

		if (meaning == EDM_INSERT_CHAR) {
			*buf = ch;
			new_x = old_x + wn_strnwidth(buf, 1, font);
		}
		else
			new_x = wn_strnwidth(s, newpos, font);
		width = wn_strwidth(s + pos, font) + edesc->ed_scrlen;

		/*  Shunt the bytes.  Note that we copy the final NUL here.
		 *  We can't use memcpy because the regions overlap.
		 *  We can't use memmove because it is not widespread enough.
		 */
		if (newpos > pos) {
			for (i = edesc->ed_curlen; i >= pos; --i)
				s[i + (newpos - pos)] = s[i];
		}
		else {
			for (i = pos; i <= edesc->ed_curlen; ++i)
				s[i - (pos - newpos)] = s[i];
		}

		wn_move_area(wn, old_x,0, width,height, new_x,0, edesc->ed_bg);
		edesc->ed_curlen = newlen;
	}

	if (meaning == EDM_INSERT_CHAR) {
		buf[0] = ch;
		buf[1] = '\0';
		wn_text(wn, font, buf, old_x, 0,
					edesc->ed_fg, edesc->ed_bg, WN_USE_TOP);
		edesc->ed_copy[pos] = ch;
	      }

	edesc->ed_curpos = newpos;
	ed_cursor(edesc, EDC_CURSOR_ON);
	wn_updating_on(wn);

	return 0;
}

/*  Turn the editing cursor on or off. Use colour INVERT so same set_area
 *  call is used for on and off. Don't need to do anything if the cursor
 *  is already off and on_or_off is off.
 *  Does block or vertical bar cursor.
 */
void
ed_cursor(edesc, cursor_on)
register struct edescst *edesc;
int cursor_on;
{
	font_t *font;
	int fg, bg, pos, ch;
	char *s, buf[2];

	if (!(edesc->ed_flags & EDF_CURSOR_ON) == cursor_on) {
		font = edesc->ed_font;
		s = edesc->ed_copy;
		pos = edesc->ed_curpos;
		if (edesc->ed_flags & EDF_BLOCK_CURSOR) {
			ch = (s[pos] != '\0') ? s[pos] : ' ';
			if (cursor_on) {
				fg = edesc->ed_bg;
				bg = edesc->ed_fg;
			}
			else {
				fg = edesc->ed_fg;
				bg = edesc->ed_bg;
			}
			buf[0] = ch;
			buf[1] = '\0';
			wn_text(edesc->ed_wn, font, buf,
					wn_strnwidth(s, pos, font), 0,
					fg, bg, WN_USE_TOP);
		}
		else	
			wn_set_area(edesc->ed_wn,
					wn_strnwidth(s, pos, font), 0,
					2, font->ft_height, INVERT);
		edesc->ed_flags ^= EDF_CURSOR_ON;
	}
}	
